/*
**  uucpd.c -- Tiny UUCP daemon which can bind to a particular address
**
**  This code is derived from mini-inetd.c of tcputils version
**  0.6.2, a collection of programs to facilitate TCP programming in
**  shell-scripts (see ftp://ftp.lysator.liu.se/pub/unix/tcputils/).
**  This package originally was released into the public domain by its
**  author <bellman@lysator.liu.se>. So this derived code is again
**  placed into the public domain, of course.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <limits.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

static int
get_inaddr(
    struct sockaddr_in *addr,
    const char *host,
    const char *service,
    const char *protocol)
{
    struct hostent *hp;
    char *end;
    long portno;
    struct servent *serv;

    memset(addr, 0, sizeof *addr);
    addr->sin_family = AF_INET;

    /* set host part of address */
    if (host == NULL)
        addr->sin_addr.s_addr = INADDR_ANY;
    else {
        addr->sin_addr.s_addr = inet_addr(host);
        if (addr->sin_addr.s_addr == (unsigned long) -1) {
            if ((hp = gethostbyname(host)) == NULL)
                return -1;
            memcpy(&addr->sin_addr, hp->h_addr, hp->h_length);
            addr->sin_family = hp->h_addrtype;
        }
    }

    /* set port part of address */
    if (service == NULL)
        addr->sin_port = htons(0);
    else {
        portno = strtol(service, &end, 10);
        if (portno > 0 && portno <= 65535 && end != service && *end == '\0')
            addr->sin_port = htons(portno);
        else {
            if ((serv = getservbyname(service, protocol)) == NULL)
                return -1;
            addr->sin_port = serv->s_port;
        }
    }
    return 0;
}

static int
tcp_listen(
    const char *interface,
    const char *port)
{
    int s;
    struct protoent *proto;
    struct sockaddr_in server;
    int yes = 1;

    if (get_inaddr(&server, interface, port, "tcp") < 0)
        return -1;
    if ((proto = getprotobyname("tcp")) == NULL)
        return -1;
    if ((s = socket(PF_INET, SOCK_STREAM, proto->p_proto)) < 0)
        return -1;
    setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&yes, sizeof(yes));
    if (bind(s, (struct sockaddr *)&server, sizeof(server)) < 0) {
        close(s);
        return -1;
    }
    if (listen(s, 256) < 0) {
        close(s);
        return -1;
    }
    return s;
}

extern char **environ;
extern int getopt(int, char *const *, const char *);
extern int optind;
extern char *optarg;

static char *progname = "uucpd";
static int debug = 0;
static int max_connections = 0;

static void
usage(
    void)
{
    fprintf(stderr,
            "Usage: %s [-d] [-m max] [-r chroot] [addr:]port argv0 [argv1 argv2 ...]\n",
            progname);
    exit(1);
}

static void
vfatal(
    const char *fmt,
    va_list args)
{
    fputs(progname, stderr);
    fputs(": ", stderr);
    vfprintf(stderr, fmt, args);
    putc('\n', stderr);
    exit(1);
}

static void
fatal(
    const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vfatal(fmt, args);
    va_end(args);
    return;
}

static void
reap_children(
    int signo)
{
    pid_t child;
    int status;
    char buf[64 + sizeof(long) * CHAR_BIT];

    while ((child = waitpid(-1, &status, WNOHANG)) > (pid_t)0) {
        if (debug) {
            sprintf(buf, "%s: child 0x%ld: terminated\n", progname, (long)child);
            write(2, buf, strlen(buf));
        }
    }
    signal(signo, &reap_children);
    return;
}

int
main(
    int argc,
    char **argv)
{
    int sd;
    pid_t child;
    int option;
    char *host;
    char *port;
    char *chrootdir = NULL;

    while ((option = getopt(argc, argv, "dm:r:")) != EOF) {
        switch ((char) option) {
            case 'd':
                debug++;
                break;
            case 'm':
                max_connections = atoi(optarg);
                break;
            case 'r':
                chrootdir = strdup(optarg);
                break;
            default:
                usage();
                /* NOTREACHED */
        }
    }
    if (argc - optind < 2)
        usage();

    if ((port = strrchr(argv[optind], ':')) != NULL) {
        *(port++) = '\0';
        host = argv[optind++];
    }
    else {
        port = argv[optind++];
        host = NULL;
    }
    if ((sd = tcp_listen(host, port)) < 0)
        fatal("Can't listen to port %s: %s", port, strerror(errno));

#if defined(SIGCHLD)
    signal(SIGCHLD, &reap_children);
#elif defined(SIGCLD)
    signal(SIGCLD, &reap_children);
#endif

    while (1) {
        struct sockaddr_in them;
        int len = sizeof them;
        int ns = accept(sd, (struct sockaddr *) &them, &len);
        if (ns < 0) {
            if (errno == EINTR)
                continue;
            fatal("Accept failed: %s", strerror(errno));
        }
        switch (child = fork()) {
            case -1:
                perror("Can't fork");
                break;
            case 0:
                /* Child */
                close(sd);
                dup2(ns, 0);
                dup2(ns, 1);
                if (chrootdir != NULL) {
                    if (chdir(chrootdir) == -1)
                        fatal("failed to chdir to %s", chrootdir);
                    if (chroot(chrootdir) == -1)
                        fatal("failed to chroot to %s", chrootdir);
                }
                execve(argv[optind], argv + optind, environ);
                fatal("Can't start %s: %s", argv[optind], strerror(errno));
            default:
                /* Parent */
                if (debug)
                    fprintf(stderr, "%s: child 0x%ld: started\n", progname, (long)child);
                close(ns);
                break;
        }
        if (max_connections > 0 && --max_connections <= 0)
            break;
    }
    return 0;
}

